from obelisk.types.core import Filter, Comparison
from datetime import datetime


def test_basic_filter():
    test_dt = datetime.now()
    f = Filter() \
        .add_and(
            Comparison.equal('source', 'test source'),
        )\
        .add_or(
            Comparison.less('timestamp', test_dt)
        )\
        .add_or(
            Comparison.is_in('metricType', ['number', 'number[]']),
        )

    expected = f"((('source'=='test source'),'timestamp'<'{test_dt.isoformat()}'),'metricType'=in=('number', 'number[]'))"
    assert str(f) == expected



"""
This module contains the asynchronous API to interface with Obelisk CORE.
These methods all return a :class:`Awaitable`.

Relevant entrance points are :class:`Client`.

This API vaguely resembles that of clients to previous Obelisk versions,
but also significantly diverts from it where the underlying Obelisk CORE API does so.
"""

from obelisk.asynchronous.base import BaseClient
from obelisk.exceptions import ObeliskError
from obelisk.types.core import FieldName, Filter

from datetime import datetime, timedelta
import httpx
import json
from pydantic import (
    BaseModel,
    AwareDatetime,
    ConfigDict,
    Field,
    ValidationError,
    model_validator,
)
from typing import (
    Annotated,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Any,
    cast,
    get_args,
)
from typing_extensions import Self
from numbers import Number

from obelisk.strategies.retry import NoRetryStrategy, RetryStrategy
from obelisk.types import ObeliskKind


DataType = Literal["number", "number[]", "json", "bool", "string"]
"""The possible types of data Obelisk can accept"""


def type_suffix(metric: str) -> DataType:
    """
    Extracts the :any:`DataType` from a string metric,
    useful for the dataType field in queries.

    Throws a :py:exc:`ValueError` if the provided string does not appear to be a typed metric,
    or the found type suffix is not a known one.
    """
    split = metric.split("::")

    if len(split) != 2:
        raise ValueError("Incorrect amount of type qualifiers")

    suffix = split[1]
    if suffix not in get_args(DataType):
        raise ValueError(
            f"Invalid type suffix, should be one of {', '.join(get_args(DataType))}"
        )
    return cast(DataType, suffix)


Aggregator = Literal["last", "min", "mean", "max", "count", "stddev"]
"""Type of aggregation Obelisk can process"""


Datapoint = Dict[str, Any]
"""Datapoints resulting from queries are modeled as simple dicts, as fields can come and go depending on query."""


class ObeliskPosition(BaseModel):
    """
    Format for coordinates as expected by Obelisk.
    """

    lat: float
    """Latitude"""
    lng: float
    """Longitude"""
    elevation: float


class IncomingDatapoint(BaseModel):
    """A datapoint to be submitted to Obelisk. These are validated quite extensively, but not fully.
    .. automethod:: check_metric_type(self)
    """

    timestamp: Optional[AwareDatetime] = None
    metric: str
    value: Any
    labels: Optional[Dict[str, str]] = None
    location: Optional[ObeliskPosition] = None

    @model_validator(mode="after")
    def check_metric_type(self) -> Self:
        suffix = type_suffix(self.metric)

        if suffix == "number" and not isinstance(self.value, Number):
            raise ValueError(
                f"Type suffix mismatch, expected number, got {type(self.value)}"
            )

        if suffix == "number[]":
            if type(self.value) is not list or any(
                [not isinstance(x, Number) for x in self.value]
            ):
                raise ValueError("Type suffix mismatch, expected value of number[]")

        # Do not check json, most things should be serialisable

        if suffix == "bool" and type(self.value) is not bool:
            raise ValueError(
                f"Type suffix mismatch, expected bool, got {type(self.value)}"
            )

        if suffix == "string" and type(self.value) is not str:
            raise ValueError(
                f"Type suffix mismatch, expected bool, got {type(self.value)}"
            )

        return self


class QueryParams(BaseModel):
    dataset: str
    groupBy: Optional[List[FieldName]] = None
    aggregator: Optional[Aggregator] = None
    fields: Optional[List[FieldName]] = None
    orderBy: Optional[List[str]] = (
        None  # More complex than just FieldName, can be prefixed with - to invert sort
    )
    dataType: Optional[DataType] = None
    filter_: Annotated[Optional[str | Filter], Field(serialization_alias="filter")] = (
        None
    )
    """
    Obelisk CORE handles filtering in `RSQL format <https://obelisk.pages.ilabt.imec.be/obelisk-core/query.html#rsql-format>`__ ,
    to make it easier to also programatically write these filters, we provide the :class:`Filter` option as well.

    Suffix to avoid collisions.
    """
    cursor: Optional[str] = None
    limit: int = 1000

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def check_datatype_needed(self) -> Self:
        if self.fields is None or "value" in self.fields:
            if self.dataType is None:
                raise ValueError("Value field requested, must specify datatype")

        return self

    def to_dict(self) -> Dict:
        return self.model_dump(exclude_none=True, by_alias=True)


class ChunkedParams(BaseModel):
    dataset: str
    groupBy: Optional[List[FieldName]] = None
    aggregator: Optional[Aggregator] = None
    fields: Optional[List[FieldName]] = None
    orderBy: Optional[List[str]] = (
        None  # More complex than just FieldName, can be prefixed with - to invert sort
    )
    dataType: Optional[DataType] = None
    filter_: Optional[str | Filter] = None
    """Underscore suffix to avoid name collisions"""
    start: datetime
    end: datetime
    jump: timedelta = timedelta(hours=1)

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def check_datatype_needed(self) -> Self:
        if self.fields is None or "value" in self.fields:
            if self.dataType is None:
                raise ValueError("Value field requested, must specify datatype")

        return self

    def chunks(self) -> Iterator[QueryParams]:
        current_start = self.start
        while current_start < self.end:
            current_end = current_start + self.jump
            filter_ = f"timestamp>={current_start.isoformat()};timestamp<{current_end.isoformat()}"
            if self.filter_:
                filter_ += f";{self.filter_}"

            yield QueryParams(
                dataset=self.dataset,
                groupBy=self.groupBy,
                aggregator=self.aggregator,
                fields=self.fields,
                orderBy=self.orderBy,
                dataType=self.dataType,
                filter_=filter_,
            )

            current_start += self.jump


class QueryResult(BaseModel):
    cursor: Optional[str] = None
    items: List[Datapoint]


class Client(BaseClient):
    page_limit: int = 250
    """How many datapoints to request per page in a cursored fetch"""

    def __init__(
        self,
        client: str,
        secret: str,
        retry_strategy: RetryStrategy = NoRetryStrategy(),
    ) -> None:
        BaseClient.__init__(
            self,
            client=client,
            secret=secret,
            retry_strategy=retry_strategy,
            kind=ObeliskKind.CORE,
        )

    async def send(
        self,
        dataset: str,
        data: List[IncomingDatapoint],
    ) -> httpx.Response:
        """
        Publishes data to Obelisk

        Parameters
        ----------
        dataset : str
            ID for the dataset to publish to
        data : List[IncomingDatapoint]
            List of Obelisk-acceptable datapoints.
            Exact format varies between Classic or HFS,
            caller is responsible for formatting.

        Raises
        ------

        ObeliskError
            When the resulting status code is not 204, an :exc:`~obelisk.exceptions.ObeliskError` is raised.
        """

        response = await self.http_post(
            f"{self.kind.root_url}/{dataset}/data/ingest",
            data=[x.model_dump(mode="json") for x in data],
        )
        if response.status_code != 204:
            msg = f"An error occured during data ingest. Status {response.status_code}, message: {response.text}"
            self.log.warning(msg)
            raise ObeliskError(msg)
        return response

    async def fetch_single_chunk(self, params: QueryParams) -> QueryResult:
        response = await self.http_get(
            f"{self.kind.root_url}/{params.dataset}/data/query", params=params.to_dict()
        )

        if response.status_code != 200:
            self.log.warning(f"Unexpected status code: {response.status_code}")
            raise ObeliskError(response.status_code, response.reason_phrase)

        try:
            js = response.json()
            return QueryResult.model_validate(js)
        except json.JSONDecodeError as e:
            msg = f"Obelisk response is not a JSON object: {e}"
            self.log.warning(msg)
            raise ObeliskError(msg)
        except ValidationError as e:
            msg = f"Response cannot be validated: {e}"
            self.log.warning(msg)
            raise ObeliskError(msg)

    async def query(self, params: QueryParams) -> List[Datapoint]:
        params.cursor = None
        result_set: List[Datapoint] = []
        result_limit = params.limit

        # Obelisk CORE does not actually stop emitting a cursor when done, limit serves as page limit
        params.limit = self.page_limit

        while True:
            result: QueryResult = await self.fetch_single_chunk(params)
            result_set.extend(result.items)
            params.cursor = result.cursor

            if len(result_set) >= result_limit or result.cursor is None:
                break

        return result_set

    async def query_time_chunked(
        self, params: ChunkedParams
    ) -> AsyncIterator[List[Datapoint]]:
        for chunk in params.chunks():
            yield await self.query(chunk)



import asyncio
from datetime import datetime, timedelta
from math import floor
from typing import Generator, List, Literal, Optional

import httpx

from obelisk.asynchronous import Obelisk as AsyncObelisk
from obelisk.strategies.retry import NoRetryStrategy, RetryStrategy
from obelisk.types import (
    Datapoint,
    IngestMode,
    ObeliskKind,
    QueryResult,
    TimestampPrecision,
)


class Obelisk:
    """
    Component that contains all the logic to consume data from
    the Obelisk API (e.g. historical data, sse).

    Wraps :class:`~obelisk.asynchronous.consumer.Consumer`.

    Obelisk API Documentation:
    https://obelisk.docs.apiary.io/
    """

    loop: asyncio.AbstractEventLoop
    """Event loop used to run interal async operations"""
    async_obelisk: AsyncObelisk
    """The actual implementation this synchronous wrapper refers to"""

    def __init__(
        self,
        client: str,
        secret: str,
        retry_strategy: RetryStrategy = NoRetryStrategy(),
        kind: ObeliskKind = ObeliskKind.CLASSIC,
    ):
        self.async_obelisk = AsyncObelisk(client, secret, retry_strategy, kind)
        self.loop = asyncio.get_event_loop()

    def fetch_single_chunk(
        self,
        datasets: List[str],
        metrics: Optional[List[str]] = None,
        fields: Optional[List[str]] = None,
        from_timestamp: Optional[int] = None,
        to_timestamp: Optional[int] = None,
        order_by: Optional[dict] = None,
        filter_: Optional[dict] = None,
        limit: Optional[int] = None,
        limit_by: Optional[dict] = None,
        cursor: Optional[str] = None,
    ) -> QueryResult:
        """
        Queries one chunk of events from Obelisk for given parameters,
        does not handle paging over Cursors.

        Parameters
        ----------

        datasets : List[str]
            List of Dataset IDs.
        metrics : Optional[List[str]] = None
            List of Metric IDs or wildcards (e.g. ``*::number``), defaults to all metrics.
        fields : Optional[List[str]] = None
            List of fields to return in the result set.
            Defaults to `[metric, source, value]`
        from_timestamp : Optional[int] = None
            Limit output to events after (and including)
            this UTC millisecond timestamp, if present.
        to_timestamp : Optional[int] = None
            Limit output to events before (and excluding)
            this UTC millisecond timestamp, if present.
        order_by : Optional[dict] = None
            Specifies the ordering of the output,
            defaults to ascending by timestamp.
            See Obelisk docs for format. Caller is responsible for validity.
        filter_ : Optional[dict] = None
            Limit output to events matching the specified Filter expression.
            See Obelisk docs, caller is responsible for validity.
        limit : Optional[int] = None
            Limit output to a maximum number of events.
            Also determines the page size.
            Default is server-determined, usually 2500.
        limit_by : Optional[dict] = None
            Limit the combination of a specific set of Index fields
            to a specified maximum number.
        cursor : Optional[str] = None
            Specifies the next cursor,
            used when paging through large result sets.
        """

        self.async_obelisk.log.info("Starting task")
        task = self.loop.create_task(
            self.async_obelisk.fetch_single_chunk(
                datasets,
                metrics,
                fields,
                from_timestamp,
                to_timestamp,
                order_by,
                filter_,
                limit,
                limit_by,
                cursor,
            )
        )
        self.async_obelisk.log.info("Blocking...")
        return self.loop.run_until_complete(task)

    def query(
        self,
        datasets: List[str],
        metrics: Optional[List[str]] = None,
        fields: Optional[List[str]] = None,
        from_timestamp: Optional[int] = None,
        to_timestamp: Optional[int] = None,
        order_by: Optional[dict] = None,
        filter_: Optional[dict] = None,
        limit: Optional[int] = None,
        limit_by: Optional[dict] = None,
    ) -> List[Datapoint]:
        """
        Queries data from obelisk,
        automatically iterating when a cursor is returned.

        Parameters
        ----------

        datasets  : List[str]
            List of Dataset IDs.
        metrics : Optional[List[str]] = None
            List of Metric IDs or wildcards (e.g. `*::number`), defaults to all metrics.
        fields : Optional[List[str]] = None
            List of fields to return in the result set.
            Defaults to `[metric, source, value]`
        from_timestamp : Optional[int] = None
            Limit output to events after (and including)
            this UTC millisecond timestamp, if present.
        to_timestamp : Optional[int] = None
            Limit output to events before (and excluding)
            this UTC millisecond timestamp, if present.
        order_by : Optional[dict] = None
            Specifies the ordering of the output,
            defaults to ascending by timestamp.
            See Obelisk docs for format. Caller is responsible for validity.
        filter_ : Optional[dict] = None
            Limit output to events matching the specified Filter expression.
            See Obelisk docs, caller is responsible for validity.
        limit : Optional[int] = None
            Limit output to a maximum number of events.
            Also determines the page size.
            Default is server-determined, usually 2500.
        limit_by : Optional[dict] = None
            Limit the combination of a specific set of Index fields
            to a specified maximum number.
        """

        task = self.loop.create_task(
            self.async_obelisk.query(
                datasets,
                metrics,
                fields,
                from_timestamp,
                to_timestamp,
                order_by,
                filter_,
                limit,
                limit_by,
            )
        )
        return self.loop.run_until_complete(task)

    def query_time_chunked(
        self,
        datasets: List[str],
        metrics: List[str],
        from_time: datetime,
        to_time: datetime,
        jump: timedelta,
        filter_: Optional[dict] = None,
        direction: Literal["asc", "desc"] = "asc",
    ) -> Generator[List[Datapoint], None, None]:
        """
        Fetches all data matching the provided filters,
        yielding one chunk at a time.
        One "chunk" may require several Obelisk calls to resolve cursors.
        By necessity iterates over time, no other ordering is supported.

        Parameters
        ----------

        datasets : List[str]
            Dataset IDs to query from
        metrics : List[str]
            IDs of metrics to query
        from_time : datetime.datetime
            Start time to fetch from
        to_time : datetime.datetime
            End time to fetch until.
        jump : datetime.timedelta
            Size of one yielded chunk
        filter_ : Optional[dict] = None
            Obelisk filter, caller is responsible for correct format
        direction : Literal['asc', 'desc'] = 'asc'
            Yield older data or newer data first, defaults to older first.
        """

        current_start = from_time
        while current_start < to_time:
            yield self.query(
                datasets=datasets,
                metrics=metrics,
                from_timestamp=floor(current_start.timestamp() * 1000),
                to_timestamp=floor((current_start + jump).timestamp() * 1000 - 1),
                order_by={"field": ["timestamp"], "ordering": direction},
                filter_=filter_,
            )
            current_start += jump

    def send(
        self,
        dataset: str,
        data: List[dict],
        precision: TimestampPrecision = TimestampPrecision.MILLISECONDS,
        mode: IngestMode = IngestMode.DEFAULT,
    ) -> httpx.Response:
        """
        Publishes data to Obelisk

        Parameters
        ----------
        dataset : str
            ID for the dataset to publish to
        data : List[dict]
            List of Obelisk-acceptable datapoints.
            Exact format varies between Classic or HFS,
            caller is responsible for formatting.
        precision : TimestampPrecision = TimestampPrecision.MILLISECONDS
            Precision used in the numeric timestamps contained in data.
            Ensure it matches to avoid weird errors.
        mode : IngestMode = IngestMode.DEFAULT
            See docs for :class:`~obelisk.types.IngestMode`.

        Raises
        ------

        ObeliskError
            When the resulting status code is not 204, an empty :exc:`~obelisk.exceptions.ObeliskError` is raised.
        """

        task = self.loop.create_task(
            self.async_obelisk.send(dataset, data, precision, mode)
        )
        return self.loop.run_until_complete(task)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
from datetime import datetime

project = 'Obelisk-Py'
copyright = '2025-{}, PreDiCT - IDLab'.format(
    datetime.now().year
)
author = 'Stef Pletinck'

import re
import os
release = re.sub('^v', '', os.popen('git describe').read().strip())

import sys
sys.path.insert(0, os.path.abspath('../..'))

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.viewcode',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
]
autosummary_generate = True
autosummary_ignore_module_all = False

templates_path = ['_templates']
exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']

autodoc_default_options = {
    'inherited-members': False,
    'undoc-members': True,
}
autoclass_content = 'class'
autodoc_member_order = 'groupwise'

intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'pydantic': ('https://docs.pydantic.dev/latest', None),
}



class AuthenticationError(Exception):
    """
    Error thrown specifically when an authentication call fails,
    usually this indicates an invalid token.
    """

    pass


class ObeliskError(Exception):
    """
    Catch-all Exception for any issue in this library code.
    """

    pass



"""
Obelisk-py is a client for the Obelisk data platform.
We support both "classic" Obelisk and HFS,
each with a synchronous and async API.
We also support Obelisk CORE, in async only for now.
The PyPi package name is ``obelisk-py``, the Python module is called ``obelisk``.

Your starting point will be one of the Obelisk instances in :mod:`~.sync` or :mod:`~.asynchronous` depending on your preferred API.

The Obelisk classes in these modules both implement the same interface,
but the asynchronous implementation returns Coroutines.

Error handling
--------------

Obelisk-py comes with robust retry logic to handle any errors that may come up.
Issues like timeouts, temporary server errors or even DNS issues are fairly common, and handling them properly is important.
Each Client accepts a retry strategy of type :class:`~.strategies.retry.RetryStrategy`.
Several predefined strategies are available in :mod:`.strategies.retry`.
"""



"""
This module contains the asynchronous API to Obelisk-py.
These methods all return an :external+python:class:`collections.abc.Awaitable`.

Relevant entrance points are :class:`client.Obelisk`.
It can be imported from the :mod:`.client` module, or directly from this one.
"""

__all__ = ["Obelisk", "core"]
from .client import Obelisk
from . import core



from obelisk.asynchronous.core import QueryParams

def test_query_param_serialize():
    q = QueryParams(dataset="83989232", filter_="(metric=='smartphone.application::string')", dataType='string')
    dump = q.to_dict()
    assert "filter" in dump
